import os  # noqa
import shutil
from pathlib import Path

import nox
from laminci import upload_docs_artifact
from laminci.nox import build_docs, login_testuser1, login_testuser2, run_pre_commit

# we'd like to aggregate coverage information across sessions
# and for this the code needs to be located in the same
# directory in every github action runner
# this also allows to break out an installation section
nox.options.default_venv_backend = "none"


GROUPS = {}
GROUPS["guide"] = [
    "tutorial.ipynb",
    "data.ipynb",
    "meta.ipynb",
    "schemas.ipynb",
    "setup.ipynb",
]
GROUPS["biology"] = [
    "bio-registries.ipynb",
]


@nox.session
def lint(session: nox.Session) -> None:
    run_pre_commit(session)


@nox.session
@nox.parametrize(
    "group",
    ["unit", "guide", "biology", "faq", "storage", "docs"],
)
def install(session, group):
    # run with pypi install on main
    if os.getenv("GITHUB_EVENT_NAME") != "push":
        # run with submodule install on a PR
        submodules = " ".join(
            [
                "./sub/lamindb-setup",
                "./sub/lnschema-core",
            ]
        )
        session.run(*f"pip install --no-deps {submodules}".split())
    extras = ""
    if group == "unit":
        extras += ",bionty,aws,zarr,postgres,fcs"
    elif group == "guide":
        extras += ",aws,bionty,zarr,jupyter,erdiagram"
        session.run(*"pip install scanpy".split())
    elif group == "biology":
        extras += ",bionty,fcs,jupyter"
    elif group == "faq":
        extras += ",aws,postgres,bionty,jupyter"
    elif group == "storage":
        extras += ",aws,zarr"
    elif group == "docs":
        extras += ",bionty"
    if os.getenv("GITHUB_EVENT_NAME") != "push":
        if "bionty" in extras:
            session.run(*"pip install --no-deps ./sub/lnschema-bionty".split())
    session.run(*f"pip install -e .[dev{extras}]".split())


@nox.session
@nox.parametrize(
    "group",
    ["unit", "guide", "biology", "faq", "storage"],
)
def build(session, group):
    login_testuser2(session)
    login_testuser1(session)
    coverage_args = "--cov=lamindb --cov-append --cov-report=term-missing"
    if group == "unit":
        session.run(*f"pytest {coverage_args} ./tests".split())
    elif group == "guide":
        session.run(
            *f"pytest -s {coverage_args} ./docs/test_notebooks.py::test_{group}".split()
        )
    elif group == "biology":
        session.run(
            *f"pytest -s {coverage_args} ./docs/test_notebooks.py::test_{group}".split()
        )
    elif group == "faq":
        session.run(*f"pytest -s {coverage_args} ./docs/faq".split())
    elif group == "storage":
        session.run(*f"pytest -s {coverage_args} ./docs/storage".split())
    # move artifacts into right place
    if group in {"guide", "biology"}:
        target_dir = Path(f"./docs/{group}")
        target_dir.mkdir(exist_ok=True)
        for filename in GROUPS[group]:
            shutil.copy(Path("docs") / filename, target_dir / filename)


@nox.session
def docs(session):
    # move artifacts into right place
    for group in ["guide", "biology", "faq", "storage"]:
        if Path(f"./docs-{group}").exists():
            if Path(f"./docs/{group}").exists():
                shutil.rmtree(f"./docs/{group}")
            Path(f"./docs-{group}").rename(f"./docs/{group}")
        # move back to root level
        if group in {"guide", "biology"}:
            for path in Path(f"./docs/{group}").glob("*"):
                path.rename(f"./docs/{path.name}")
    login_testuser1(session)
    session.run(*"lamin init --storage ./docsbuild --schema bionty".split())
    build_docs(session, strip_prefix=True, strict=True)
    upload_docs_artifact(aws=True)



import pytest

from lamindb_setup.dev._hub_core import add_storage


@pytest.fixture(scope="session")
def s3_bucket_1():
    return "lndb-setup-ci"


def test_add_storage(user_account_1, auth_1, s3_bucket_1):
    storage_id, message = add_storage(
        root=f"s3://{s3_bucket_1}",
        account_handle=user_account_1["handle"],
        _access_token=auth_1["access_token"],
    )

    assert storage_id
    assert message is None


@pytest.mark.skip("This test assumes we have AWS credentials set")
def test_add_storage_with_non_existing_bucket(auth_1):
    non_existing_storage_root = "s3://non_existing_storage_root"

    storage_id, message = add_storage(
        root=non_existing_storage_root,
        account_handle=auth_1["handle"],
        _access_token=auth_1["access_token"],
    )
    assert storage_id is None
    assert message == "bucket-does-not-exists"



from inspect import signature

import lnschema_bionty as lb
import pandas as pd
import pytest
from lnschema_core.models import FileLabel
from pandas.api.types import is_categorical_dtype, is_string_dtype

import lamindb as ln
from lamindb import _feature

df = pd.DataFrame(
    {
        "feat1": [1, 2, 3],
        "feat2": [3.1, 4.2, 5.3],
        "feat3": ["cond1", "cond2", "cond2"],
        "feat4": ["id1", "id2", "id3"],
        "rando_feature": ["rando1", "rando2", "rando3"],
    }
)


def test_signatures():
    # this seems currently the easiest and most transparent
    # way to test violations of the signature equality
    # the MockORM class is needed to get inspect.signature
    # to work
    class Mock:
        pass

    # class methods
    class_methods = ["from_df"]
    for name in class_methods:
        setattr(Mock, name, getattr(_feature, name))
        assert signature(getattr(Mock, name)) == _feature.SIGS.pop(name)
    # methods
    for name, sig in _feature.SIGS.items():
        assert signature(getattr(_feature, name)) == sig


def test_feature_from_df():
    # try to generate the file without validated features
    file = ln.File.from_df(df, description="test")
    assert file._feature_sets == {}
    # now, register all 4 features
    ln.save(ln.Feature.from_df(df.iloc[:, :4]))
    # try again
    file = ln.File.from_df(df, description="test")
    assert "columns" in file._feature_sets
    file.save()
    feature_set = file._feature_sets["columns"]
    features = feature_set.features.all()
    assert len(features) == len(df.columns[:4])
    string_cols = [col for col in df.columns if is_string_dtype(df[col])]
    categoricals = {col: df[col] for col in df.columns if is_categorical_dtype(df[col])}
    for key in string_cols:
        c = pd.Categorical(df[key])
        if len(c.categories) < len(c):
            categoricals[key] = c
    for feature in features:
        if feature.name in categoricals:
            assert feature.type == "category"
        else:
            orig_type = df[feature.name].dtype.name
            orig_type_stripped = "".join(i for i in orig_type if not i.isdigit())
            assert feature.type == orig_type_stripped
    for feature in features:
        feature.save()
    labels = ln.Label.from_values(df["feat3"])
    ln.save(labels)
    file.add_labels(labels, feature="feat3")
    assert set(ln.Label.filter(filelabel__feature__name="feat3").list("name")) == set(
        ["cond1", "cond2"]
    )
    for name in df.columns[:4]:
        queried_feature = ln.Feature.filter(name=name).one()
        if name in categoricals:
            assert queried_feature.type == "category"
        else:
            orig_type = df[name].dtype.name
            orig_type_stripped = "".join(i for i in orig_type if not i.isdigit())
            assert queried_feature.type == orig_type_stripped
    filelabel_links = FileLabel.objects.filter(file_id=file.id, feature__name="feat3")
    label_ids = filelabel_links.values_list("label_id")
    assert set(
        ln.Label.objects.filter(id__in=label_ids).values_list("name", flat=True)
    ) == set(["cond1", "cond2"])
    for feature in features:
        feature.delete()
    file.delete(storage=True)


def test_feature_init():
    # no args allowed
    with pytest.raises(ValueError):
        ln.Feature("x")
    # registries has to be a list of Registry types
    with pytest.raises(ValueError):
        ln.Feature(name="feat", type="category", registries=1)
    # each element of the list has to be a Registry
    with pytest.raises(ValueError):
        ln.Feature(name="feat", type="category", registries=[1])
    # registries_str
    feature = ln.Feature(name="feat1", type="category", registries=[ln.Label, lb.Gene])
    assert feature.registries == "core.Label|bionty.Gene"



"""Open-source data platform for biology.

LaminDB helps you manage data using registries.
The most central are:

.. autosummary::
   :toctree: .

   File
   Dataset

There are four registries to track provenance of data batches:

.. autosummary::
   :toctree: .

   Transform
   Run
   User
   Storage

And four registries to validate & contextualize measurements in data batches:

.. autosummary::
   :toctree: .

   Label
   Feature
   FeatureSet
   Modality

Functional tools:

.. autosummary::
   :toctree: .

   track
   view
   save

Static classes & modules:

.. autosummary::
   :toctree: .

   settings
   types
   setup
   schema
   dev

"""

__version__ = "0.51a1"  # denote a release candidate for 0.1.0 with 0.1rc1

import os as _os

import lamindb_setup as _lamindb_setup

# prints warning of python versions
from lamin_utils import py_version_warning as _py_version_warning
from lamindb_setup import _check_instance_setup
from lamindb_setup._check_instance_setup import _INSTANCE_NOT_SETUP_WARNING

_py_version_warning("3.8", "3.11")

_TESTING = _lamindb_setup._TESTING
_INSTANCE_SETUP = _check_instance_setup(from_lamindb=True)
# allow the user to call setup
from . import setup  # noqa


class InstanceNotSetupError(Exception):
    pass


def __getattr__(name):
    raise InstanceNotSetupError(
        f"{_INSTANCE_NOT_SETUP_WARNING}If you used the CLI to init or load an instance,"
        " please RESTART the python session (in a notebook, restart kernel)"
    )


# only import all other functionality if setup was successful
if _INSTANCE_SETUP:
    del InstanceNotSetupError
    del __getattr__  # delete so that imports work out
    from lnschema_core import (  # noqa
        Dataset,
        Feature,
        FeatureSet,
        File,
        Label,
        Modality,
        Run,
        Storage,
        Transform,
        User,
    )

    from . import dev  # noqa
    from . import schema  # noqa
    from . import types  # noqa
    from .dev._run_context import run_context  # noqa

    track = run_context._track  # noqa
    from lamin_utils import logger as _logger

    from . import _dataset  # noqa
    from . import _feature  # noqa
    from . import _feature_set  # noqa
    from . import _file  # noqa
    from . import _label  # noqa
    from . import _parents  # noqa
    from . import _registry  # noqa
    from . import _storage  # noqa
    from . import _synonym  # noqa
    from . import _transform  # noqa
    from . import _validate  # noqa
    from ._delete import delete  # noqa
    from ._registry import select_backward as select  # noqa
    from ._save import save  # noqa
    from ._view import view  # noqa
    from .dev._settings import settings

    add = save  # backward compat

    settings.__doc__ = """Global :class:`~lamindb.dev.Settings`."""
